from ..model import BasicTaggedToken


class TaggedToken(BasicTaggedToken):
    """
    Tagged token with token types.

    Attributes:
        tokenTypeSet (set): Token types for this token.
    """

    def __init__(self, tokenTypeSet=None):
        super().__init__()
        self.tokenTypeSet = tokenTypeSet

    def clearTokenType(self):
        if not self.tokenTypeSet:
            self.tokenTypeSet.clear()



from .model import TimeSegment
from .util import NlpUtil
from .util import RegexUtil
from .util import TokenTypeUtil


class SynTime(object):
    """
    SynTime APIs.
    """

    def __init__(self):
        self.__nlpUtil = NlpUtil()
        self.__regexUtil = RegexUtil()
        self.__tokenTypeUtil = TokenTypeUtil(self.__regexUtil)

    def extractTimexFromText(self, text, date):
        """
        Extract timex (time expressions) from text.

        Args:
            text (str): Raw text to extract timex from.
            date (str)
        """
        if not text:
            return text
        taggedTokenList = self.__nlpUtil.tagging(text)
        timeTokenList = self.__identifyTimeToken(taggedTokenList)
        if not timeTokenList or len(timeTokenList) == 0:
            return text
        timeSegmentList = self.__identifyTimeSegment(taggedTokenList, timeTokenList)
        if not timeSegmentList or len(timeSegmentList) == 0:
            return text
        return self.__generateTimeMLText(text, taggedTokenList, timeSegmentList, date)

    def __identifyTimeToken(self, taggedTokenList):
        """
        Get time token list from taggedTokenList.

        Args:
            taggedTokenList (list): Tagged token list.

        Returns:
            timeTokenList (list): Time token index list.
        """
        timeTokenList = []
        for i in range(len(taggedTokenList)):
            taggedToken = taggedTokenList[i]
            token = taggedToken.token
            tag = taggedToken.tag

            tokenTypeSet = self.__tokenTypeUtil.getTokenTypeSet(token, tag)
            taggedToken.tokenTypeSet = tokenTypeSet

            if TokenTypeUtil.isHalfDayToken(taggedToken):
                if token.lower() == 'am' or token.lower() == 'pm':
                    if i == 0 or (not TokenTypeUtil.isNumeralToken(taggedTokenList[i - 1])
                                  and not TokenTypeUtil.isTimeToken(taggedTokenList[i - 1])):
                        taggedToken.clearTokenType()
            elif TokenTypeUtil.isTimeZoneToken(taggedToken):
                if i == 0 or not TokenTypeUtil.isTimeToken(taggedTokenList[i - 1]) \
                        and not TokenTypeUtil.isHalfDayToken(taggedTokenList[i - 1]) \
                        and not TokenTypeUtil.isNumeralToken(taggedTokenList[i - 1]):
                    taggedToken.clearTokenType()
            elif TokenTypeUtil.isEraToken(taggedToken):
                if i == 0 or not TokenTypeUtil.isNumeralToken(taggedTokenList[i - 1]):
                    taggedToken.clearTokenType()
            elif TokenTypeUtil.isYearToken(taggedToken):
                if i > 0 and (TokenTypeUtil.isInArticleToken(taggedTokenList[i - 1])
                              or taggedTokenList[i - 1].token.lower() == 'the'):
                    taggedTokenList[i - 1].clearTokenType()

            if TokenTypeUtil.isGeneralTimeToken(taggedToken):
                timeTokenList.append(i)

        return timeTokenList

    def __identifyTimeSegment(self, taggedTokenList, timeTokenList):
        """
        Get time segment list from taggedTokenList and timeTokenList.

        Args:
            taggedTokenList (list): Tagged token list.
            timeTokenList (list): Time token position list.

        Returns:
            timeSegmentList (list): Time segment list.
        """
        timeSegmentList = []
        firstToken = 0
        lastToken = len(taggedTokenList) - 1
        timeTokenLen = len(timeTokenList)
        for i in range(timeTokenLen):
            timeTokenPosition = timeTokenList[i]

            beginToken = timeTokenPosition
            endToken = timeTokenPosition

            taggedTimeToken = taggedTokenList[timeTokenPosition]
            if TokenTypeUtil.isPeriodToken(taggedTimeToken) or TokenTypeUtil.isDurationToken(taggedTimeToken):
                timeSegmentList.append(TimeSegment(timeTokenPosition, beginToken, endToken))
                continue

            leftBound = firstToken
            rightBound = lastToken
            if i > 0:
                leftBound = timeTokenList[i - 1] + 1
            if i < timeTokenLen - 1:
                rightBound = timeTokenList[i + 1] - 1

            # Search its left side
            findLeftDependentSegment = False
            leftTimeTokenPosition = -1
            leftBeginToken = leftTimeTokenPosition
            leftEndToken = leftTimeTokenPosition
            j = timeTokenPosition - 1
            while j >= leftBound:
                taggedPreMod = taggedTokenList[j]
                if TokenTypeUtil.isPrefixToken(taggedPreMod) or TokenTypeUtil.isNumeralToken(taggedPreMod) \
                        or TokenTypeUtil.isInArticleToken(taggedPreMod):
                    beginToken = j
                elif TokenTypeUtil.isCommaToken(taggedPreMod):
                    beginToken = j
                    break
                elif TokenTypeUtil.isLinkageToken(taggedPreMod):
                    if j - 1 >= leftBound and TokenTypeUtil.isGeneralTimeToken(taggedTimeToken) \
                            and TokenTypeUtil.isNumeralToken(taggedTokenList[j - 1]):
                        findLeftDependentSegment = True
                        leftTimeTokenPosition = j - 1
                        leftBeginToken = leftTimeTokenPosition
                        leftEndToken = leftTimeTokenPosition
                        k = leftTimeTokenPosition - 1
                        while k >= leftBound:
                            if TokenTypeUtil.isPrefixToken(taggedPreMod) or TokenTypeUtil.isNumeralToken(taggedPreMod) \
                                    or TokenTypeUtil.isInArticleToken(taggedPreMod):
                                leftBeginToken = k
                                k -= 1
                            else:
                                break
                        beginToken = j + 1
                    else:
                        beginToken = j
                    break
                else:
                    break
                j -= 1

            # Search its right side
            findRightDependentSegment = False
            rightTimeTokenPosition = 0
            rightBeginToken = rightTimeTokenPosition
            rightEndToken = rightTimeTokenPosition
            j = timeTokenPosition + 1
            while j <= rightBound:
                taggedSufMod = taggedTokenList[j]
                if TokenTypeUtil.isSuffixToken(taggedSufMod) or TokenTypeUtil.isNumeralToken(taggedSufMod):
                    endToken = j
                elif TokenTypeUtil.isCommaToken(taggedSufMod):
                    endToken = j
                    break
                elif TokenTypeUtil.isLinkageToken(taggedSufMod):
                    if j + 1 <= rightBound and TokenTypeUtil.isGeneralTimeToken(taggedTimeToken) \
                            and TokenTypeUtil.isNumeralToken(taggedTokenList[j + 1]):
                        findRightDependentSegment = True
                        rightTimeTokenPosition = j + 1
                        rightBeginToken = rightTimeTokenPosition
                        rightEndToken = rightTimeTokenPosition
                        k = rightTimeTokenPosition + 1
                        while k <= rightBound:
                            if TokenTypeUtil.isSuffixToken(taggedSufMod) or TokenTypeUtil.isNumeralToken(taggedSufMod):
                                rightEndToken = k
                                k += 1
                            else:
                                break
                        endToken = j - 1
                    else:
                        endToken = j
                    break
                else:
                    break
                j += 1

            if findLeftDependentSegment:
                timeSegmentList.append(TimeSegment(leftTimeTokenPosition, leftBeginToken, leftEndToken))
            timeSegmentList.append(TimeSegment(timeTokenPosition, beginToken, endToken))
            if findRightDependentSegment:
                timeSegmentList.append(TimeSegment(rightTimeTokenPosition, rightBeginToken, rightEndToken))

        return timeSegmentList

    def __generateTimeMLText(self, text, taggedTokenList, timeSegmentList, date):
        """
        Generate TimeML text for input text.

        Args:
            text (str): Raw text.
            taggedTokenList (list): Tagged token list.
            timeSegmentList (list): Time segment list.
            date (str): User-specified date.

        Returns:
            timeMLText (str): Text in TimeML format.
        """
        timeMLText = ''
        type = 'DATE'
        value = date
        tid = 1
        lastCharPosition = 0
        taggedTokenListLen = len(taggedTokenList)
        timeSegmentListLen = len(timeSegmentList)

        if timeSegmentList and timeSegmentListLen > 0:
            isTimex = True
            timexBeginTokenPosition = timeSegmentList[0].beginTokenPosition
            timexEndTokenPosition = timeSegmentList[0].endTokenPosition

            for i in range(timeSegmentListLen):
                timeSegment = timeSegmentList[i]
                segmentBeginTokenPosition = timeSegment.beginTokenPosition
                segmentEndTokenPosition = timeSegment.endTokenPosition
                if timexEndTokenPosition + 1 == segmentBeginTokenPosition \
                        or timexEndTokenPosition > segmentBeginTokenPosition:
                    isTimex = False
                elif timexEndTokenPosition == segmentBeginTokenPosition:
                    if TokenTypeUtil.isCommaToken(taggedTokenList[segmentBeginTokenPosition]):
                        if segmentBeginTokenPosition == 0 or segmentBeginTokenPosition + 1 == taggedTokenListLen:
                            isTimex = True
                        else:
                            commaPreToken = taggedTokenList[segmentBeginTokenPosition - 1]
                            commaSufToken = taggedTokenList[segmentBeginTokenPosition + 1]
                            if (TokenTypeUtil.isGeneralTimeToken(commaPreToken) or TokenTypeUtil.isNumeralToken(
                                    commaPreToken)) and TokenTypeUtil.isGeneralTimeToken(commaSufToken) \
                                    and not TokenTypeUtil.isSameTokenType(commaPreToken, commaSufToken):
                                isTimex = False
                            else:
                                isTimex = True
                    elif TokenTypeUtil.isLinkageToken(taggedTokenList[segmentBeginTokenPosition]):
                        isTimex = True
                    else:
                        isTimex = False
                else:
                    isTimex = True

                if not isTimex:
                    timexEndTokenPosition = segmentEndTokenPosition
                else:
                    timexBeginTaggedToken = taggedTokenList[timexBeginTokenPosition]
                    timexEndTaggedToken = taggedTokenList[timexEndTokenPosition]
                    if TokenTypeUtil.isCommaToken(timexBeginTaggedToken) or TokenTypeUtil.isLinkageToken(
                            timexBeginTaggedToken) or timexBeginTaggedToken.tag == 'IN':
                        timexBeginTokenPosition += 1
                    if TokenTypeUtil.isComma(timexEndTaggedToken) or TokenTypeUtil.isLinkageToken(timexEndTaggedToken):
                        timexEndTokenPosition -= 1

                    timexEndToken = timexEndTaggedToken.token
                    beginCharPosition = timexBeginTaggedToken.beginCharPosition
                    endCharPosition = timexEndTaggedToken.endCharPosition

                    timeMLText += text[lastCharPosition:beginCharPosition]
                    lastCharPosition = beginCharPosition

                    index = timexBeginTokenPosition
                    while index <= timexEndTokenPosition:
                        temTaggedToken = taggedTokenList[index]
                        if TokenTypeUtil.isYearYearToken(temTaggedToken) or TokenTypeUtil.isYearMonthToken(
                                temTaggedToken) \
                                or TokenTypeUtil.isMonthMonthToken(temTaggedToken) or TokenTypeUtil.isWeekWeekToken(
                            temTaggedToken) \
                                or TokenTypeUtil.isTimeTimeToken(temTaggedToken) or TokenTypeUtil.isHalfDayHalfDayToken(
                            temTaggedToken) \
                                or TokenTypeUtil.isNumeralNumeralToken(temTaggedToken):
                            temBeginCharPosition = temTaggedToken.beginCharPosition
                            items = temTaggedToken.token.split('-')
                            timex = text[lastCharPosition:temBeginCharPosition + len(items[0])]
                            timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex) + '-'
                            lastCharPosition = temBeginCharPosition + len(items[0]) + 1
                            tid += 1
                        index += 1

                    if timexEndToken.endswith('\'s'):
                        timex = text[lastCharPosition:endCharPosition - 2]
                        timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex)
                        lastCharPosition = endCharPosition - 2
                    elif timexEndToken.endswith('s') \
                            and (timexEndTokenPosition + 1 < taggedTokenListLen and
                                 taggedTokenList[timexEndTokenPosition + 1].token == "'"):
                        timex = text[lastCharPosition:taggedTokenList[timexEndTokenPosition + 1].endCharPosition]
                        timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex)
                        lastCharPosition = taggedTokenList[timexEndTokenPosition + 1].endCharPosition
                    else:
                        timex = text[lastCharPosition:endCharPosition]
                        timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex)
                        lastCharPosition = endCharPosition
                    tid += 1
                    timexBeginTokenPosition = segmentBeginTokenPosition
                    timexEndTokenPosition = segmentEndTokenPosition

            timexBeginTaggedToken = taggedTokenList[timexBeginTokenPosition]
            timexEndTaggedToken = taggedTokenList[timexEndTokenPosition]
            if TokenTypeUtil.isCommaToken(timexBeginTaggedToken) or TokenTypeUtil.isLinkageToken(
                    timexBeginTaggedToken) or timexBeginTaggedToken.tag == 'IN':
                timexBeginTokenPosition += 1
            if TokenTypeUtil.isComma(timexEndTaggedToken) or TokenTypeUtil.isLinkageToken(timexEndTaggedToken):
                timexEndTokenPosition -= 1

            timexEndToken = timexEndTaggedToken.token
            beginCharPosition = timexBeginTaggedToken.beginCharPosition
            endCharPosition = timexEndTaggedToken.endCharPosition

            timeMLText += text[lastCharPosition:beginCharPosition]
            lastCharPosition = beginCharPosition

            index = timexBeginTokenPosition
            while index <= timexEndTokenPosition:
                temTaggedToken = taggedTokenList[index]
                if TokenTypeUtil.isYearYearToken(temTaggedToken) or TokenTypeUtil.isYearMonthToken(
                        temTaggedToken) \
                        or TokenTypeUtil.isMonthMonthToken(temTaggedToken) or TokenTypeUtil.isWeekWeekToken(
                    temTaggedToken) \
                        or TokenTypeUtil.isTimeTimeToken(temTaggedToken) or TokenTypeUtil.isHalfDayHalfDayToken(
                    temTaggedToken) \
                        or TokenTypeUtil.isNumeralNumeralToken(temTaggedToken):
                    temBeginCharPosition = temTaggedToken.beginCharPosition
                    items = temTaggedToken.token.split('-')
                    timex = text[lastCharPosition:temBeginCharPosition + len(items[0])]
                    timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex) + '-'
                    lastCharPosition = temBeginCharPosition + len(items[0]) + 1
                    tid += 1
                index += 1

            if timexEndToken.endswith('\'s'):
                timex = text[lastCharPosition:endCharPosition - 2]
                timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex)
                lastCharPosition = endCharPosition - 2
            elif timexEndToken.endswith('s') \
                    and (timexEndTokenPosition + 1 < taggedTokenListLen and
                         taggedTokenList[timexEndTokenPosition + 1].token == "'"):
                timex = text[lastCharPosition:taggedTokenList[timexEndTokenPosition + 1].endCharPosition]
                timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex)
                lastCharPosition = taggedTokenList[timexEndTokenPosition + 1].endCharPosition
            else:
                timex = text[lastCharPosition:endCharPosition]
                timeMLText += SynTime.__getTIMEX3Str(tid, type, value, timex)
                lastCharPosition = endCharPosition
            tid += 1

        timeMLText += text[lastCharPosition:]
        return timeMLText

    @classmethod
    def __getTIMEX3Str(cls, tid, timexType, value, timex):
        TIMEX3_TID = "<TIMEX3 tid=\"t"
        TIMEX3_TYPE = "\" type=\""
        TIMEX3_VALUE = "\" value=\""
        TIMEX3_MID = "\">"
        TIMEX3_END = "</TIMEX3>"
        return TIMEX3_TID + str(tid) + TIMEX3_TYPE + timexType + TIMEX3_VALUE + value + TIMEX3_MID + timex + TIMEX3_END



class TimeSegment(object):
    """
    Time segment.

    Attributes:
        timeTokenPosition (int): Time token position in the text.
        beginTokenPosition (int): Begin position in the text.
        endTokenPosition (int): End position in the text.
    """

    def __init__(self, timeTokenPosition, beginTokenPosition, endTokenPosition):
        self.timeTokenPosition = timeTokenPosition
        self.beginTokenPosition = beginTokenPosition
        self.endTokenPosition = endTokenPosition



from ..model import TokenType
from ..model import TaggedToken


class TokenTypeUtil(object):

    def __init__(self, regexUtil):
        self.regexUtil = regexUtil

    def getTokenTypeSet(self, token, posTag):
        """
        Get token type set for specific token with specific pos-tag.

        Args:
            token (str)
            posTag (str)
        """
        tokenTypeSet = set()
        isPosTagNN = posTag.startswith('NN')
        isPosTagJJ = posTag == 'JJ'
        isPosTagCD = posTag == 'CD'
        isPosTagRB = posTag == 'RB'

        if self.regexUtil.YEAR_PATTERN_1.fullmatch(token) or self.regexUtil.YEAR_PATTERN_2.fullmatch(token) \
                or self.regexUtil.YEAR_MID_PATTERN.fullmatch(token) or self.regexUtil.ERA_YEAR_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.YEAR)

        mYearYear = self.regexUtil.YEAR_YEAR_PATTERN.fullmatch(token)
        if mYearYear:
            tokenTypeSet.add(TokenType.YEAR_YEAR)

        if isPosTagNN and self.regexUtil.SEASON_PATTERN.fullmatch(token) \
                or self.regexUtil.SEASON_MID_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.SEASON)

        if isPosTagNN and self.regexUtil.MONTH_PATTERN.fullmatch(token) \
                or self.regexUtil.MONTH_MID_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.MONTH)
        if isPosTagNN and self.regexUtil.MONTH_ABBR_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.MONTH_ABBR)
        if self.regexUtil.MONTH_MONTH_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.MONTH_MONTH)

        if self.regexUtil.YEAR_MONTH_PATTERN_1.fullmatch(token) or self.regexUtil.YEAR_MONTH_PATTERN_2.fullmatch(token):
            tokenTypeSet.add(TokenType.YEAR_MONTH)

        if isPosTagNN and self.regexUtil.WEEK_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.WEEK)

        if isPosTagNN and self.regexUtil.WEEK_ABBR_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.WEEK_ABBR)

        if self.regexUtil.WEEK_WEEK_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.WEEK_WEEK)

        if self.regexUtil.DATE_PATTERN_1.fullmatch(token) or self.regexUtil.DATE_PATTERN_2.fullmatch(token) \
                or self.regexUtil.DATE_PATTERN_3.fullmatch(token):
            tokenTypeSet.add(TokenType.DATE)

        if self.regexUtil.TIME_PATTERN_1.fullmatch(token) or self.regexUtil.TIME_PATTERN_2.fullmatch(token):
            tokenTypeSet.add(TokenType.TIME)
        if self.regexUtil.TIME_TIME_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.TIME_TIME)

        if self.regexUtil.HALFDAY_PATTERN_1.fullmatch(token) or self.regexUtil.HALFDAY_PATTERN_2.fullmatch(token):
            tokenTypeSet.add(TokenType.HALFDAY)
        if self.regexUtil.HALFDAY_HALFDAY_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.HALFDAY_HALFDAY)

        if self.regexUtil.TIME_ZONE_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.TIME_ZONE)

        if isPosTagNN and self.regexUtil.ERA_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.ERA)

        if isPosTagNN and self.regexUtil.TIME_UNIT_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.TIME_UNIT)

        mDuration = self.regexUtil.DURATION_PATTERN.fullmatch(token)
        if (isPosTagNN or isPosTagJJ or isPosTagCD) and mDuration:
            tokenTypeSet.add(TokenType.DURATION)

        if not mDuration and (
                self.regexUtil.DURATION_DURATION_PATTERN_1.fullmatch(token)
                or self.regexUtil.DURATION_DURATION_PATTERN_2.fullmatch(token)):
            tokenTypeSet.add(TokenType.DURATION_DURATION)

        if (isPosTagNN or isPosTagRB or isPosTagJJ or self.regexUtil.DAY_TIME_MID_PATTERN.fullmatch(token)) and \
                self.regexUtil.DAY_TIME_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.DAY_TIME)

        if (isPosTagNN or isPosTagRB) and self.regexUtil.TIMELINE_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.TIMELINE)

        if self.regexUtil.HOLIDAY_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.HOLIDAY)

        if (isPosTagNN or isPosTagRB or isPosTagJJ) and self.regexUtil.PERIOD_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.PERIOD)

        if isPosTagNN and self.regexUtil.DECADE_PATTERN.fullmatch(token) or self.regexUtil.DECADE_MID_PATTERN.fullmatch(
                token):
            tokenTypeSet.add(TokenType.DECADE)

        mBasicNum2 = self.regexUtil.BASIC_NUMBER_PATTERN_2.fullmatch(token)
        if ((isPosTagJJ or isPosTagCD or isPosTagRB) and (self.regexUtil.ORDINAL_PATTERN_1.fullmatch(token)
                                                          or self.regexUtil.ORDINAL_PATTERN_2.fullmatch(token))) \
                or self.regexUtil.DIGIT_PATTERN_1.fullmatch(token) \
                or self.regexUtil.DIGIT_PATTERN_2.fullmatch(token) \
                or self.regexUtil.BASIC_NUMBER_PATTERN_1.fullmatch(token) \
                or mBasicNum2:
            tokenTypeSet.add(TokenType.NUMERAL)

        if not mYearYear and not mBasicNum2 and (
                self.regexUtil.DIGIT_DIGIT_PATTERN.fullmatch(
                    token) or self.regexUtil.BASIC_NUMBER_NUMBER_PATTERN.fullmatch(token)
                or self.regexUtil.ORDINAL_ORDINAL_PATTERN.fullmatch(token)):
            tokenTypeSet.add(TokenType.NUMERAL_NUMERAL)

        if self.regexUtil.INARTICLE_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.INARTICLE)

        if len(tokenTypeSet) == 0 and (not isPosTagNN and self.regexUtil.PREFIX_PATTERN_1.fullmatch(token) or
                                     isPosTagNN and self.regexUtil.PREFIX_PATTERN_2.fullmatch(token)):
            tokenTypeSet.add(TokenType.PREFIX)

        if self.regexUtil.SUFFIX_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.SUFFIX)

        if self.regexUtil.LINKAGE_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.LINKAGE)

        if self.regexUtil.COMMA_PATTERN.fullmatch(token):
            tokenTypeSet.add(TokenType.COMMA)

        return tokenTypeSet

    @classmethod
    def hasType(cls, tokenTypeSet, targetTypeFunc):
        """
        Check if tokenTypeSet contains a target type.
        Args:
            tokenTypeSet (set): Token type set.
            targetTypeFunc (function): Function to determine if the tokenTypeSet contain a target type.
        """
        for tokenType in tokenTypeSet:
            if targetTypeFunc(tokenType):
                return True
        return False

    @classmethod
    def isYear(cls, tokenType):
        """
        Check if the token type is year.
        Args:
            tokenType (TaggedToken): Token type.
        """
        return tokenType == TokenType.YEAR or tokenType == TokenType.YEAR_YEAR

    @classmethod
    def hasYear(cls, tokenTypeSet):
        """
        Check if one of the token types is year.
        Args:
            tokenTypeSet (set): Token type set.
        """
        return cls.hasType(tokenTypeSet, cls.isYear)

    @classmethod
    def isYearToken(cls, taggedToken):
        """
        Check if taggedToken has year token type in its tokenTypeSet.
        Args:
            taggedToken (TaggedToken): Tagged token.
        """
        return cls.hasYear(taggedToken.tokenTypeSet)

    @classmethod
    def isYearYear(cls, tokenType):
        return tokenType == TokenType.YEAR_YEAR

    @classmethod
    def hasYearYear(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isYearYear)

    @classmethod
    def isYearYearToken(cls, taggedToken):
        return cls.hasYearYear(taggedToken.tokenTypeSet)

    @classmethod
    def isSeason(cls, tokenType):
        return tokenType == TokenType.SEASON

    @classmethod
    def hasSeason(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isSeason)

    @classmethod
    def isSeasonToken(cls, taggedToken):
        return cls.hasSeason(taggedToken.tokenTypeSet)

    @classmethod
    def isMonth(cls, tokenType):
        return tokenType == TokenType.MONTH or tokenType == TokenType.MONTH_ABBR \
               or tokenType == TokenType.MONTH_MONTH or tokenType == TokenType.YEAR_MONTH

    @classmethod
    def hasMonth(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isMonth)

    @classmethod
    def isMonthToken(cls, taggedToken):
        return cls.hasMonth(taggedToken.tokenTypeSet)

    @classmethod
    def isMonthMonth(cls, tokenType):
        return tokenType == TokenType.MONTH_MONTH

    @classmethod
    def hasMonthMonth(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isMonthMonth)

    @classmethod
    def isMonthMonthToken(cls, taggedToken):
        return cls.hasMonthMonth(taggedToken.tokenTypeSet)

    @classmethod
    def isYearMonth(cls, tokenType):
        return tokenType == TokenType.YEAR_MONTH

    @classmethod
    def hasYearMonth(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isYearMonth)

    @classmethod
    def isYearMonthToken(cls, taggedToken):
        return cls.hasYearMonth(taggedToken.tokenTypeSet)

    @classmethod
    def isWeek(cls, tokenType):
        return tokenType == TokenType.WEEK or tokenType == TokenType.WEEK_ABBR or tokenType == TokenType.WEEK_WEEK

    @classmethod
    def hasWeek(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isWeek)

    @classmethod
    def isWeekToken(cls, taggedToken):
        return cls.hasWeek(taggedToken.tokenTypeSet)

    @classmethod
    def isWeekWeek(cls, tokenType):
        return tokenType == TokenType.WEEK_WEEK

    @classmethod
    def hasWeekWeek(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isWeekWeek)

    @classmethod
    def isWeekWeekToken(cls, taggedToken):
        return cls.hasWeekWeek(taggedToken.tokenTypeSet)

    @classmethod
    def isDate(cls, tokenType):
        return tokenType == TokenType.DATE

    @classmethod
    def hasDate(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isDate)

    @classmethod
    def isDateToken(cls, taggedToken):
        return cls.hasDate(taggedToken.tokenTypeSet)

    @classmethod
    def isTime(cls, tokenType):
        return tokenType == TokenType.TIME or tokenType == TokenType.TIME_TIME

    @classmethod
    def hasTime(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isTime)

    @classmethod
    def isTimeToken(cls, taggedToken):
        return cls.hasTime(taggedToken.tokenTypeSet)

    @classmethod
    def isTimeTime(cls, tokenType):
        return tokenType == TokenType.TIME_TIME

    @classmethod
    def hasTimeTime(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isTimeTime)

    @classmethod
    def isTimeTimeToken(cls, taggedToken):
        return cls.hasTimeTime(taggedToken.tokenTypeSet)

    @classmethod
    def isHalfDay(cls, tokenType):
        return tokenType == TokenType.HALFDAY or tokenType == TokenType.HALFDAY_HALFDAY

    @classmethod
    def hasHalfDay(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isHalfDay)

    @classmethod
    def isHalfDayToken(cls, taggedToken):
        return cls.hasHalfDay(taggedToken.tokenTypeSet)

    @classmethod
    def isHalfDayHalfDay(cls, tokenType):
        return tokenType == TokenType.HALFDAY_HALFDAY

    @classmethod
    def hasHalfDayHalfDay(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isHalfDayHalfDay)

    @classmethod
    def isHalfDayHalfDayToken(cls, taggedToken):
        return cls.hasHalfDayHalfDay(taggedToken.tokenTypeSet)

    @classmethod
    def isTimeZone(cls, tokenType):
        return tokenType == TokenType.TIME_ZONE

    @classmethod
    def hasTimeZone(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isTimeZone)

    @classmethod
    def isTimeZoneToken(cls, taggedToken):
        return cls.hasTimeZone(taggedToken.tokenTypeSet)

    @classmethod
    def isEra(cls, tokenType):
        return tokenType == TokenType.ERA

    @classmethod
    def hasEra(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isEra)

    @classmethod
    def isEraToken(cls, taggedToken):
        return cls.hasEra(taggedToken.tokenTypeSet)

    @classmethod
    def isTimeUnit(cls, tokenType):
        return tokenType == TokenType.TIME_UNIT

    @classmethod
    def hasTimeUnit(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isTimeUnit)

    @classmethod
    def isTimeUnitToken(cls, taggedToken):
        return cls.hasTimeUnit(taggedToken.tokenTypeSet)

    @classmethod
    def isDuration(cls, tokenType):
        return tokenType == TokenType.DURATION

    @classmethod
    def hasDuration(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isDuration)

    @classmethod
    def isDurationToken(cls, taggedToken):
        return cls.hasDuration(taggedToken.tokenTypeSet)

    @classmethod
    def isDurationDuration(cls, tokenType):
        return tokenType == TokenType.DURATION_DURATION

    @classmethod
    def hasDurationDuration(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isDurationDuration)

    @classmethod
    def isDurationDurationToken(cls, taggedToken):
        return cls.hasDurationDuration(taggedToken.tokenTypeSet)

    @classmethod
    def isDayTime(cls, tokenType):
        return tokenType == TokenType.DAY_TIME

    @classmethod
    def hasDayTime(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isDayTime)

    @classmethod
    def isDayTimeToken(cls, taggedToken):
        return cls.hasDayTime(taggedToken.tokenTypeSet)

    @classmethod
    def isTimeLine(cls, tokenType):
        return tokenType == TokenType.TIMELINE

    @classmethod
    def hasTimeLine(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isTimeLine)

    @classmethod
    def isTimeLineToken(cls, taggedToken):
        return cls.hasTimeLine(taggedToken.tokenTypeSet)

    @classmethod
    def isHoliday(cls, tokenType):
        return tokenType == TokenType.HOLIDAY

    @classmethod
    def hasHoliday(cls, tokenType):
        return cls.hasType(tokenType, cls.isHoliday)

    @classmethod
    def isHolidayToken(cls, taggedToken):
        return cls.hasHoliday(taggedToken.tokenTypeSet)

    @classmethod
    def isPeriod(cls, tokenType):
        return tokenType == TokenType.PERIOD

    @classmethod
    def hasPeriod(cls, tokenType):
        return cls.hasType(tokenType, cls.isPeriod)

    @classmethod
    def isPeriodToken(cls, taggedToken):
        return cls.hasPeriod(taggedToken.tokenTypeSet)

    @classmethod
    def isDecade(cls, tokenType):
        return tokenType == TokenType.DECADE

    @classmethod
    def hasDecade(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isDecade)

    @classmethod
    def isDecadeToken(cls, taggedToken):
        return cls.hasDecade(taggedToken.tokenTypeSet)

    @classmethod
    def isNumeral(cls, tokenType):
        return tokenType == TokenType.NUMERAL or tokenType == TokenType.NUMERAL_NUMERAL

    @classmethod
    def hasNumeral(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isNumeral)

    @classmethod
    def isNumeralToken(cls, taggedToken):
        return cls.hasNumeral(taggedToken.tokenTypeSet)

    @classmethod
    def isNumeralNumeral(cls, tokenType):
        return tokenType == TokenType.NUMERAL_NUMERAL

    @classmethod
    def hasNumeralNumeral(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isNumeralNumeral)

    @classmethod
    def isNumeralNumeralToken(cls, taggedToken):
        return cls.hasNumeralNumeral(taggedToken.tokenTypeSet)

    @classmethod
    def isInArticle(cls, tokenType):
        return tokenType == TokenType.INARTICLE

    @classmethod
    def hasInArticle(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isInArticle)

    @classmethod
    def isInArticleToken(cls, taggedToken):
        return cls.hasInArticle(taggedToken.tokenTypeSet)

    @classmethod
    def isPrefix(cls, tokenType):
        return tokenType == TokenType.PREFIX

    @classmethod
    def hasPrefix(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isPrefix)

    @classmethod
    def isPrefixToken(cls, taggedToken):
        return cls.hasPrefix(taggedToken.tokenTypeSet)

    @classmethod
    def isSuffix(cls, tokenType):
        return tokenType == TokenType.SUFFIX

    @classmethod
    def hasSuffix(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isSuffix)

    @classmethod
    def isSuffixToken(cls, taggedToken):
        return cls.hasSuffix(taggedToken.tokenTypeSet)

    @classmethod
    def isLinkage(cls, tokenType):
        return tokenType == TokenType.LINKAGE

    @classmethod
    def hasLinkage(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isLinkage)

    @classmethod
    def isLinkageToken(cls, taggedToken):
        return cls.hasLinkage(taggedToken.tokenTypeSet)

    @classmethod
    def isComma(cls, tokenType):
        return tokenType == TokenType.COMMA

    @classmethod
    def hasComma(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isComma)

    @classmethod
    def isCommaToken(cls, taggedToken):
        return cls.hasComma(taggedToken.tokenTypeSet)

    @classmethod
    def isGeneralTime(cls, tokenType):
        return cls.isYear(tokenType) or cls.isSeason(tokenType) or cls.isMonth(tokenType) or cls.isWeek(
            tokenType) or cls.isDate(
            tokenType) or cls.isTime(tokenType) or cls.isHalfDay(tokenType) \
               or cls.isTimeZone(tokenType) or cls.isEra(tokenType) or cls.isTimeUnit(tokenType) or cls.isDuration(
            tokenType) or cls.isDayTime(tokenType) or cls.isTimeLine(tokenType) \
               or cls.isHoliday(tokenType) or cls.isPeriod(tokenType) or cls.isDecade(tokenType)

    @classmethod
    def hasGeneralTime(cls, tokenTypeSet):
        return cls.hasType(tokenTypeSet, cls.isGeneralTime)

    @classmethod
    def isGeneralTimeToken(cls, taggedToken):
        return cls.hasGeneralTime(taggedToken.tokenTypeSet)

    @classmethod
    def isSameTokenType(cls, taggedToken1, taggedToken2):
        tokenTypeSet1 = taggedToken1.tokenTypeSet
        tokenTypeSet2 = taggedToken2.tokenTypeSet
        if not tokenTypeSet1 or not tokenTypeSet2 or len(tokenTypeSet1) != len(tokenTypeSet2):
            return False
        for tokenType1 in tokenTypeSet1:
            for tokenType2 in tokenTypeSet2:
                if tokenType1 == tokenType2:
                    return True
        return False



class BasicTaggedToken(object):
    """
    Basic tagged token for a token.

    Attributes:
        token (str): Raw token.
        lemma (str): Lemma.
        tag (str): POS-tag.
        tokenPosition (int): Index of the token in the text beginning from 0.
        beginCharPosition (int): Starting character position in the text.
        endCharPosition (int): Ending character position in the text.
    """

    def __init__(self):
        self.token = ""
        self.lemma = ""
        self.tag = ""
        self.tokenPosition = 0
        self.beginCharPosition = 0
        self.endCharPosition = 0

    # def __init__(self, token, lemma, tag, tokenPosition, beginCharPosition, endCharPosition):
    #     self.token = token
    #     self.lemma = lemma
    #     self.tag = tag
    #     self.tokenPosition = tokenPosition
    #     self.beginCharPosition = beginCharPosition
    #     self.endCharPosition = endCharPosition

    def __str__(self):
        return "{}, {}, {}, {}, {}, {}".format(self.token, self.lemma, self.tag, self.tokenPosition,
                                               self.beginCharPosition, self.endCharPosition)



from enum import Enum


class TokenType(Enum):
    """
    Token type enum.
    """

    YEAR = 0
    YEAR_MID = 1
    YEAR_YEAR = 2
    MONTH = 3
    MONTH_ABBR = 4
    MONTH_MID = 5
    MONTH_MONTH = 6
    YEAR_MONTH = 7
    WEEK = 8
    WEEK_ABBR = 9
    WEEK_WEEK = 10
    DATE = 11
    TIME = 12
    TIME_TIME = 13
    DAY_TIME = 14
    SEASON = 15
    TIMELINE = 16
    PERIOD = 17
    DECADE = 18
    HALFDAY = 19
    HALFDAY_HALFDAY = 20
    TIME_UNIT = 21
    DURATION = 22
    DURATION_DURATION = 23
    HOLIDAY = 24
    TIME_ZONE = 25
    ERA = 26
    PREFIX = 27
    SUFFIX = 28
    LINKAGE = 29
    COMMA = 30
    INARTICLE = 31
    BASIC_NUMBER = 32
    BASIC_NUMBER_NUMBER = 33
    ORDINAL = 34
    ORDINAL_ORDINAL = 35
    DIGIT = 36
    DIGIT_DIGIT = 37
    NUMERAL = 38
    NUMERAL_NUMERAL = 39



import spacy
from ..model import TaggedToken

class NlpUtil(object):
    """
    NLP util to do POS-tagging for the input text. NLP is powered by spaCy.
    """

    def __init__(self):
        self.nlp = spacy.load('en_core_web_sm')

    def tagging(self, text):
        if not text:
            return None
        doc = self.nlp(text)
        taggedTokenList = []
        offset = 0
        for token in doc:
            offset = text.find(token.text, offset)
            taggedToken = TaggedToken()
            taggedToken.token = token.text
            taggedToken.lemma = token.lemma_
            taggedToken.tag = token.tag_
            taggedToken.tokenPosition = token.i
            taggedToken.beginCharPosition = offset
            taggedToken.endCharPosition = offset + len(token)
            offset += len(token)
            taggedTokenList.append(taggedToken)

        return taggedTokenList



import os.path
import re


class RegexUtil(object):
    """
    Util to load and compile regex patterns for SynTime.
    """

    SYN_TIME_REGEX_FILE = os.path.join(os.path.dirname(__file__), '../resource/SynTimeRegex.txt')

    # From StanfordCoreNLP edu.stanford.nlp.ie.regexp.NumberSequenceClassifier.java
    MID_REGEX = None

    # POS: CD; Type: DATE
    YEAR_REGEX_1 = None
    YEAR_PATTERN_1 = None
    YEAR_REGEX_2 = None
    YEAR_PATTERN_2 = None
    YEAR_MID_REGEX = None
    YEAR_MID_PATTERN = None

    YEAR_YEAR_REGEX = None
    YEAR_YEAR_PATTERN = None

    # POS: NNP; Type: DATE
    MONTH_REGEX = None
    MONTH_PATTERN = None
    MONTH_ABBR_REGEX = None
    MONTH_ABBR_PATTERN = None
    MONTH_MID_REGEX = None
    MONTH_MID_PATTERN = None

    MONTH_MONTH_REGEX = None
    MONTH_MONTH_PATTERN = None

    YEAR_MONTH_REGEX_1 = None
    YEAR_MONTH_PATTERN_1 = None
    YEAR_MONTH_REGEX_2 = None
    YEAR_MONTH_PATTERN_2 = None

    # POS: NNP; Type: DATE
    WEEK_REGEX = None
    WEEK_PATTERN = None
    WEEK_ABBR_REGEX = None
    WEEK_ABBR_PATTERN = None

    WEEK_WEEK_REGEX = None
    WEEK_WEEK_PATTERN = None

    # POS: CD
    BASIC_NUMBER_REGEX_1 = None
    BASIC_NUMBER_PATTERN_1 = None
    BASIC_NUMBER_REGEX_2 = None
    BASIC_NUMBER_PATTERN_2 = None

    BASIC_NUMBER_NUMBER_REGEX = None
    BASIC_NUMBER_NUMBER_PATTERN = None

    # POS: CD
    DIGIT_REGEX_1 = None
    DIGIT_PATTERN_1 = None
    DIGIT_REGEX_2 = None
    DIGIT_PATTERN_2 = None

    DIGIT_DIGIT_REGEX = None
    DIGIT_DIGIT_PATTERN = None

    # POS: JJ, CD
    ORDINAL_REGEX_1 = None
    ORDINAL_PATTERN_1 = None
    ORDINAL_REGEX_2 = None
    ORDINAL_PATTERN_2 = None

    ORDINAL_ORDINAL_REGEX = None
    ORDINAL_ORDINAL_PATTERN = None

    INARTICLE_REGEX = None
    INARTICLE_PATTERN = None

    #  Type: DATE
    DATE_REGEX_1 = None
    DATE_PATTERN_1 = None
    DATE_REGEX_2 = None
    DATE_PATTERN_2 = None
    DATE_REGEX_3 = None
    DATE_PATTERN_3 = None

    #  Type: TIME
    TIME_REGEX_1 = None
    TIME_PATTERN_1 = None
    TIME_REGEX_2 = None
    TIME_PATTERN_2 = None
    TIME_TIME_REGEX = None
    TIME_TIME_PATTERN = None

    TIME_ZONE_REGEX = None
    TIME_ZONE_PATTERN = None

    ERA_REGEX = None
    ERA_PATTERN = None
    ERA_YEAR_REGEX = None
    ERA_YEAR_PATTERN = None

    # POS: NN, VBD; Type: TIME
    HALFDAY_REGEX_1 = None
    HALFDAY_PATTERN_1 = None

    HALFDAY_REGEX_2 = None
    HALFDAY_PATTERN_2 = None

    HALFDAY_HALFDAY_REGEX = None
    HALFDAY_HALFDAY_PATTERN = None

    # From english.sutime.txt
    # https://github.com/stanfordnlp/CoreNLP/tree/master/src/edu/stanford/nlp/time/rules
    #
    # POS: NNS; Type: DATE
    DECADE_REGEX = None
    DECADE_PATTERN = None
    DECADE_MID_REGEX = None
    DECADE_MID_PATTERN = None

    # POS: NN, NNS, NNP; Type: DURATION
    TIME_UNIT_REGEX = None
    TIME_UNIT_PATTERN = None

    # POS: NN, JJ; Type: DURATION
    DURATION_REGEX = None
    DURATION_PATTERN = None

    DURATION_DURATION_REGEX_1 = None
    DURATION_DURATION_PATTERN_1 = None
    DURATION_DURATION_REGEX_2 = None
    DURATION_DURATION_PATTERN_2 = None

    # POS: RB, JJ; Type: SET
    PERIOD_REGEX = None
    PERIOD_PATTERN = None

    # POS: NN, NNP, NNS, RB, JJ; Type: TIME
    DAY_TIME_REGEX = None
    DAY_TIME_PATTERN = None
    DAY_TIME_MID_REGEX = None
    DAY_TIME_MID_PATTERN = None

    # POS: NN, NNS; Type: DATE
    SEASON_REGEX = None
    SEASON_PATTERN = None
    SEASON_MID_REGEX = None
    SEASON_MID_PATTERN = None

    # POS: RB, NN; Type: DATE
    TIMELINE_REGEX = None
    TIMELINE_PATTERN = None

    """
    Five kinds of semantic modifiers: relative modifiers, frequency modifiers, early late modifiers, approximate modifiers, and operators.
    Relative modifier: the, next, following, last, previous, this, coming, past
    Frequency modifier: each, every, other, alternate, alternating
    Early late modifier: late, early, mid-?, beginning, start, dawn, middle, end, of, in, on
    Approximate modifier: about, around, some, exactly, precisely
    Operator: this, next, following, previous, last, this, the, coming, following, next, past, previous
    """
    # POS: non-NN*
    PREFIX_REGEX_1 = None
    PREFIX_PATTERN_1 = None
    # POS: NN*
    PREFIX_REGEX_2 = None
    PREFIX_PATTERN_2 = None

    SUFFIX_REGEX = None
    SUFFIX_PATTERN = None

    LINKAGE_REGEX = None
    LINKAGE_PATTERN = None

    COMMA_REGEX = None
    COMMA_PATTERN = None

    # From english.holidays.sutime.txt
    HOLIDAY_REGEX = None
    HOLIDAY_PATTERN = None

    def __init__(self, synTimeRegexFile=SYN_TIME_REGEX_FILE):
        self.__loadSynTimeRegex(synTimeRegexFile)
        self.__inducePattern()

    def __loadSynTimeRegex(self, synTimeRegexFile):
        with open(synTimeRegexFile) as f:
            lines = f.readlines()
        assert lines is not None, 'Failed to load SynTime regex'
        for line in lines:
            line = line.strip()
            if not line:
                continue
            regexName, regex = line.split('\t')
            regexName = regexName[1:-1]
            if regexName == 'MID_REGEX':
                self.MID_REGEX = regex
            elif regexName == 'YEAR_REGEX_1':
                self.YEAR_REGEX_1 = regex
            elif regexName == 'YEAR_REGEX_2':
                self.YEAR_REGEX_2 = regex
            elif regexName == 'MONTH_REGEX':
                self.MONTH_REGEX = regex
            elif regexName == 'MONTH_ABBR_REGEX':
                self.MONTH_ABBR_REGEX = regex
            elif regexName == 'WEEK_REGEX':
                self.WEEK_REGEX = regex
            elif regexName == 'WEEK_ABBR_REGEX':
                self.WEEK_ABBR_REGEX = regex
            elif regexName == 'BASIC_NUMBER_REGEX_1':
                self.BASIC_NUMBER_REGEX_1 = regex
            elif regexName == 'BASIC_NUMBER_REGEX_2':
                self.BASIC_NUMBER_REGEX_2 = regex
            elif regexName == 'DIGIT_REGEX_1':
                self.DIGIT_REGEX_1 = regex
            elif regexName == 'ORDINAL_REGEX_1':
                self.ORDINAL_REGEX_1 = regex
            elif regexName == 'ORDINAL_REGEX_2':
                self.ORDINAL_REGEX_2 = regex
            elif regexName == 'INARTICLE_REGEX':
                self.INARTICLE_REGEX = regex
            elif regexName == 'DATE_REGEX_1':
                self.DATE_REGEX_1 = regex
            elif regexName == 'DATE_REGEX_2':
                self.DATE_REGEX_2 = regex
            elif regexName == 'DATE_REGEX_3':
                self.DATE_REGEX_3 = regex
            elif regexName == 'TIME_REGEX_1':
                self.TIME_REGEX_1 = regex
            elif regexName == 'TIME_REGEX_2':
                self.TIME_REGEX_2 = regex
            elif regexName == 'TIME_ZONE_REGEX':
                self.TIME_ZONE_REGEX = regex
            elif regexName == 'HALFDAY_REGEX_1':
                self.HALFDAY_REGEX_1 = regex
            elif regexName == 'ERA_REGEX':
                self.ERA_REGEX = regex
            elif regexName == 'DECADE_REGEX':
                self.DECADE_REGEX = regex
            elif regexName == 'TIME_UNIT_REGEX':
                self.TIME_UNIT_REGEX = regex
            elif regexName == 'PERIOD_REGEX':
                self.PERIOD_REGEX = regex
            elif regexName == 'DAY_TIME_REGEX':
                self.DAY_TIME_REGEX = regex
            elif regexName == 'SEASON_REGEX':
                self.SEASON_REGEX = regex
            elif regexName == 'TIMELINE_REGEX':
                self.TIMELINE_REGEX = regex
            elif regexName == 'PREFIX_REGEX_1':
                self.PREFIX_REGEX_1 = regex
            elif regexName == 'PREFIX_REGEX_2':
                self.PREFIX_REGEX_2 = regex
            elif regexName == 'SUFFIX_REGEX':
                self.SUFFIX_REGEX = regex
            elif regexName == 'LINKAGE_REGEX':
                self.LINKAGE_REGEX = regex
            elif regexName == 'COMMA_REGEX':
                self.COMMA_REGEX = regex
            elif regexName == 'HOLIDAY_REGEX':
                self.HOLIDAY_REGEX = regex

    def __inducePattern(self):
        # POS: CD; Type: DATE
        self.YEAR_PATTERN_1 = re.compile(self.YEAR_REGEX_1, re.IGNORECASE)
        self.YEAR_PATTERN_2 = re.compile(self.YEAR_REGEX_2, re.IGNORECASE)
        self.YEAR_MID_REGEX = "(" + self.MID_REGEX + ")(" + self.YEAR_REGEX_1 + "|" + self.YEAR_REGEX_2 + ")"
        self.YEAR_MID_PATTERN = re.compile(self.YEAR_MID_REGEX, re.IGNORECASE)

        self.YEAR_YEAR_REGEX = self.YEAR_REGEX_1 + "-(" + self.YEAR_REGEX_1 + "|[0-9]{2})"
        self.YEAR_YEAR_PATTERN = re.compile(self.YEAR_YEAR_REGEX, re.IGNORECASE)

        # POS: NNP; Type: DATE
        self.MONTH_PATTERN = re.compile(self.MONTH_REGEX, re.IGNORECASE)
        self.MONTH_ABBR_PATTERN = re.compile(self.MONTH_ABBR_REGEX, re.IGNORECASE)
        self.MONTH_MID_REGEX = "(" + self.MID_REGEX + ")(" + self.MONTH_REGEX + "|" + self.MONTH_ABBR_REGEX + ")"
        self.MONTH_MID_PATTERN = re.compile(self.MONTH_MID_REGEX, re.IGNORECASE)

        self.MONTH_MONTH_REGEX = "(" + self.MONTH_REGEX + "|" + self.MONTH_ABBR_REGEX + ")-(" + self.MONTH_REGEX + "|" + self.MONTH_ABBR_REGEX + ")"
        self.MONTH_MONTH_PATTERN = re.compile(self.MONTH_MONTH_REGEX, re.IGNORECASE)

        self.YEAR_MONTH_REGEX_1 = self.YEAR_REGEX_1 + "-(" + self.MONTH_REGEX + "|" + self.MONTH_ABBR_REGEX + ")"
        self.YEAR_MONTH_PATTERN_1 = re.compile(self.YEAR_MONTH_REGEX_1, re.IGNORECASE)
        self.YEAR_MONTH_REGEX_2 = "(" + self.MONTH_REGEX + "|" + self.MONTH_ABBR_REGEX + ")-" + self.YEAR_REGEX_1
        self.YEAR_MONTH_PATTERN_2 = re.compile(self.YEAR_MONTH_REGEX_2, re.IGNORECASE)

        # POS: NNP; Type: DATE
        self.WEEK_PATTERN = re.compile(self.WEEK_REGEX, re.IGNORECASE)
        self.WEEK_ABBR_PATTERN = re.compile(self.WEEK_ABBR_REGEX, re.IGNORECASE)

        self.WEEK_WEEK_REGEX = "(" + self.WEEK_REGEX + "|" + self.WEEK_ABBR_REGEX + ")-(" + self.WEEK_REGEX + "|" + self.WEEK_ABBR_REGEX + ")"
        self.WEEK_WEEK_PATTERN = re.compile(self.WEEK_WEEK_REGEX, re.IGNORECASE)

        # POS: CD
        self.BASIC_NUMBER_PATTERN_1 = re.compile(self.BASIC_NUMBER_REGEX_1, re.IGNORECASE)
        self.BASIC_NUMBER_PATTERN_2 = re.compile(self.BASIC_NUMBER_REGEX_2, re.IGNORECASE)

        self.BASIC_NUMBER_NUMBER_REGEX = "(" + self.BASIC_NUMBER_REGEX_1 + ")-(" + self.BASIC_NUMBER_REGEX_1 + ")"
        self.BASIC_NUMBER_NUMBER_PATTERN = re.compile(self.BASIC_NUMBER_NUMBER_REGEX, re.IGNORECASE)

        # POS: CD
        self.DIGIT_PATTERN_1 = re.compile(self.DIGIT_REGEX_1, re.IGNORECASE)
        self.DIGIT_REGEX_2 = self.DIGIT_REGEX_1 + "[/\\.]" + self.DIGIT_REGEX_1
        self.DIGIT_PATTERN_2 = re.compile(self.DIGIT_REGEX_2, re.IGNORECASE)

        self.DIGIT_DIGIT_REGEX = self.DIGIT_REGEX_1 + "-" + self.DIGIT_REGEX_1
        self.DIGIT_DIGIT_PATTERN = re.compile(self.DIGIT_DIGIT_REGEX)

        # POS: JJ, CD
        self.ORDINAL_PATTERN_1 = re.compile(self.ORDINAL_REGEX_1, re.IGNORECASE)
        self.ORDINAL_PATTERN_2 = re.compile(self.ORDINAL_REGEX_2, re.IGNORECASE)

        self.ORDINAL_ORDINAL_REGEX = "(" + self.ORDINAL_REGEX_1 + ")-(" + self.ORDINAL_REGEX_1 + ")"
        self.ORDINAL_ORDINAL_PATTERN = re.compile(self.ORDINAL_ORDINAL_REGEX, re.IGNORECASE)

        self.INARTICLE_PATTERN = re.compile(self.INARTICLE_REGEX, re.IGNORECASE)

        #  Type: DATE
        self.DATE_PATTERN_1 = re.compile(self.DATE_REGEX_1)
        self.DATE_PATTERN_2 = re.compile(self.DATE_REGEX_2)
        self.DATE_PATTERN_3 = re.compile(self.DATE_REGEX_3, re.IGNORECASE)

        #  Type: TIME
        self.TIME_PATTERN_1 = re.compile(self.TIME_REGEX_1)
        self.TIME_PATTERN_2 = re.compile(self.TIME_REGEX_2)
        self.TIME_TIME_REGEX = "(" + self.TIME_REGEX_1 + "|" + self.TIME_REGEX_2 + ")-(" + self.TIME_REGEX_1 + "|" + self.TIME_REGEX_2 + ")"
        self.TIME_TIME_PATTERN = re.compile(self.TIME_TIME_REGEX)

        self.TIME_ZONE_PATTERN = re.compile(self.TIME_ZONE_REGEX, re.IGNORECASE)

        self.ERA_PATTERN = re.compile(self.ERA_REGEX, re.IGNORECASE)
        self.ERA_YEAR_REGEX = self.DIGIT_REGEX_1 + "(" + self.ERA_REGEX + ")"
        self.ERA_YEAR_PATTERN = re.compile(self.ERA_YEAR_REGEX, re.IGNORECASE)

        # POS: NN, VBD; Type: TIME
        self.HALFDAY_PATTERN_1 = re.compile(self.HALFDAY_REGEX_1, re.IGNORECASE)

        self.HALFDAY_REGEX_2 = "(" + self.DIGIT_REGEX_1 + "|" + self.DIGIT_REGEX_2 + "|" + self.TIME_REGEX_1 + "|" + self.TIME_REGEX_2 + ")(" + self.HALFDAY_REGEX_1 + ")"
        self.HALFDAY_PATTERN_2 = re.compile(self.HALFDAY_REGEX_2, re.IGNORECASE)

        self.HALFDAY_HALFDAY_REGEX = self.HALFDAY_REGEX_2 + "-" + self.HALFDAY_REGEX_2
        self.HALFDAY_HALFDAY_PATTERN = re.compile(self.HALFDAY_HALFDAY_REGEX, re.IGNORECASE)

        # From english.sutime.txt
        # POS: NNS; Type: DATE
        self.DECADE_PATTERN = re.compile(self.DECADE_REGEX, re.IGNORECASE)
        self.DECADE_MID_REGEX = "(" + self.MID_REGEX + ")(" + self.DECADE_REGEX + ")"
        self.DECADE_MID_PATTERN = re.compile(self.DECADE_MID_REGEX, re.IGNORECASE)

        # POS: NN, NNS, NNP; Type: DURATION
        self.TIME_UNIT_PATTERN = re.compile(self.TIME_UNIT_REGEX, re.IGNORECASE)

        # POS: NN, JJ; Type: DURATION
        self.DURATION_REGEX = "(" + self.DIGIT_REGEX_1 + "|" + self.DIGIT_REGEX_2 + "|" + self.BASIC_NUMBER_REGEX_1 + "|" + self.BASIC_NUMBER_REGEX_2 + "|" + self.ORDINAL_REGEX_1 + "|" + self.ORDINAL_REGEX_2 + "|" + self.INARTICLE_REGEX + ")-?(" + self.TIME_UNIT_REGEX + ")"
        self.DURATION_PATTERN = re.compile(self.DURATION_REGEX, re.IGNORECASE)

        self.DURATION_DURATION_REGEX_1 = "(" + self.DIGIT_DIGIT_REGEX + "|" + self.BASIC_NUMBER_NUMBER_REGEX + "|" + self.ORDINAL_ORDINAL_REGEX + ")-?(" + self.TIME_UNIT_REGEX + ")"
        self.DURATION_DURATION_PATTERN_1 = re.compile(self.DURATION_DURATION_REGEX_1, re.IGNORECASE)
        self.DURATION_DURATION_REGEX_2 = self.DURATION_REGEX + "-" + self.DURATION_REGEX
        self.DURATION_DURATION_PATTERN_2 = re.compile(self.DURATION_DURATION_REGEX_2, re.IGNORECASE)

        # POS: RB, JJ; Type: SET
        self.PERIOD_PATTERN = re.compile(self.PERIOD_REGEX, re.IGNORECASE)

        # POS: NN, NNP, NNS, RB, JJ; Type: TIME
        self.DAY_TIME_PATTERN = re.compile(self.DAY_TIME_REGEX, re.IGNORECASE)
        self.DAY_TIME_MID_REGEX = "(" + self.MID_REGEX + ")(" + self.DAY_TIME_REGEX + ")"
        self.DAY_TIME_MID_PATTERN = re.compile(self.DAY_TIME_MID_REGEX, re.IGNORECASE)

        # POS: NN, NNS; Type: DATE
        self.SEASON_PATTERN = re.compile(self.SEASON_REGEX, re.IGNORECASE)
        self.SEASON_MID_REGEX = "(" + self.MID_REGEX + ")(" + self.SEASON_REGEX + ")"
        self.SEASON_MID_PATTERN = re.compile(self.SEASON_MID_REGEX, re.IGNORECASE)

        # POS: RB, NN; Type: DATE
        self.TIMELINE_PATTERN = re.compile(self.TIMELINE_REGEX, re.IGNORECASE)

        # Modifiers
        # POS: non-NN*
        self.PREFIX_PATTERN_1 = re.compile(self.PREFIX_REGEX_1, re.IGNORECASE)
        # POS: NN*
        self.PREFIX_PATTERN_2 = re.compile(self.PREFIX_REGEX_2, re.IGNORECASE)

        self.SUFFIX_PATTERN = re.compile(self.SUFFIX_REGEX, re.IGNORECASE)

        self.LINKAGE_PATTERN = re.compile(self.LINKAGE_REGEX, re.IGNORECASE)

        self.COMMA_PATTERN = re.compile(self.COMMA_REGEX)

        # From english.holidays.sutime.txt
        self.HOLIDAY_PATTERN = re.compile(self.HOLIDAY_REGEX, re.IGNORECASE)



import re
from setuptools import setup, find_packages

with open('pysyntime/__init__.py', 'r') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        f.read(), re.MULTILINE).group(1)

with open('README.md') as f:
    readme = f.read()

with open('requirements.txt') as f:
    requirements = f.readlines()

setup(
    name='pysyntime',
    version=version,
    description='SynTime implementation in Python.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/ZhouShengsheng/pysyntime',
    author='Zhou Shengsheng',
    author_email='szhou006@e.ntu.edu.sg',
    # license='MIT',
    classifiers=[
        # 'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
)
